# A basic box parameterized tensor is represented as a triple 
# (ap, abounds, aeqs), where
# ap is the polynomial in q which is the coefficient
# abounds is a list of 4-tuples of integers (a,b,c,d), one for each place of the tensor
#       the integers, representing the corresponding bound
#       a*q + b <= i_j <= c*q + d
# aeqs is a list of pairs of place indices, indicating the equality relations
#       defining the index set

# A box parameterized tensor is represented as a list of its summands, each a
# basic box parameterized tensor

from itertools import *
from copy import deepcopy

# tensor product of two basic box parameterized tensors
def tensor_product_basic(a,b):
    ap, abounds, aeqs = a
    bp, bbounds, beqs = b
    return (ap*bp, abounds + bbounds, 
            aeqs + [(i+len(abounds),j+len(abounds)) for i,j in beqs])

# sums the basic box parameterized tensor across place i, returning the result
# and a q lower bound for which the result is valid
def sum_basic(a, i):
    ap, abounds, aeqs = deepcopy(a)
    qlb = 0
    if any((j==i and k!=i) or (k==i and j!=i) for j,k in aeqs):
        try:
            j = next(k for j,k in aeqs if j == i and k != i)
        except StopIteration:
            j = next(j for j,k in aeqs if k == i and j != i)
        qlb = 0
        if abounds[i][0] != abounds[j][0]:
            qlb = max(qlb, abs(abounds[i][1] - abounds[j][1]) )
        if abounds[i][2] != abounds[j][2]:
            qlb = max(qlb, abs(abounds[i][3] - abounds[j][3]) )

        # print i,j,abounds[i][:2], abounds[j][:2], abounds[i][2:], abounds[j][2:]
        abounds[j] = max(abounds[i][:2], abounds[j][:2]) + min(abounds[i][2:], abounds[j][2:])

        if abounds[j][:2] > abounds[j][2:]:
            return None,qlb

        aeqs.extend([(l,j) for k,l in aeqs if k==i])
        aeqs.extend([(k,j) for k,l in aeqs if l==i])

    else:
        q = QQ['q'].gen()
        a,b,c,d = abounds[i]
        ap *= c*q+d - a*q-b+1

    return (ap, abounds[:i]+abounds[i+1:], 
            [(j if j<i else j-1,k if k<i else k-1) 
                for j,k in aeqs if j!=i and k!=i]),qlb

# contracts places i and j for the basic box parameterized tensor a, returning
# the result and a q lower bound for which the result is valid
def contract_basic(a,i,j):
    a = (a[0],a[1],copy(a[2]))
    a[2].append((i,j))
    a,qlb1 = sum_basic(a,max(i,j))
    if a is None:
        return None,qlb1
    a,qlb2 = sum_basic(a,min(i,j))
    return a,max(qlb1,qlb2)

# rearanges the places of the bb tensor a according to the permutation ixs
def transpose_basic(a,ixs):
    ap,abounds,aeqs = a
    return (ap, [abounds[ixs[i]] for i in range(len(abounds))], 
            [(ixs.index(i),ixs.index(j)) for i,j in aeqs])

# tensor product of box parameterized tensors A and B
def tensor_product_box(A,B):
    return [tensor_product_basic(a,b) for a in A for b in B]

# contraction of box parameterized tensors A between places i and j. Prints the
# q lower bound for which the result is valid
def contract_box(A,i,j):
    out = []
    qlb = 0
    for a in A:
        a,lo = contract_basic(a,i,j) 
        if a is not None:
            out.append(a)
        qlb = max(qlb,lo)
    print 'q >=',qlb
    # assert contract_explicit(get_explicit(A,max(qlb,2)),i,j) == get_explicit(out,max(qlb,2))
    # aa = contract_explicit(get_explicit(A,max(qlb,2)),i,j)
    # bb = get_explicit(out,max(qlb,2))
    # if not aa == bb:
    #     import IPython
    #     IPython.embed()
    #     raise ValueError
    return out

# rearranges the places of A according to the permutation ixs
def transpose_box(A,ixs):
    return [transpose_basic(a,ixs) for a in A]

# convenience function to construct the basic box parameterized tensor
# with the polynomial e in the entry corresponding to indices ixs, and zero
# elsewhere
def constant_basic(ixs, e):
    return (e, [(0,i,0,i) for i in ixs], [])

# returns the box parameterized tensor sum_{i=0}^{n-1} a_i \ot b_i \ot ... \ot z_i, 
# where z is the kth letter
def identity_tensor(n,k=2):
    return [(1, [(0,0,0,n-1)]*k, zip(range(k-1),range(1,k)) ) ]

# expands the box paramerized tensor for fixed value of q 
# (optionally not substituting q in the polynomials)
# The result is represented sparsely as a dictionary mapping indices to values
def get_explicit(A,q,sub=True):
    T = {}
    for ap, abounds, aeqs in A:
        if sub:
            try:
                val = ap(q)
            except TypeError:
                val = ap
        else:
            val = ap
        if val == 0:
            continue
        eq = range(len(abounds))
        def find(i):
            if eq[i] != i:
                eq[i] = find(eq[i])
            return eq[i]
        def union(i,j):
            eq[find(i)] = find(j)
        for i,j in aeqs:
            union(i,j)
        bds = {}
        for i,bd in enumerate(abounds):
            a,b,c,d = bd
            lo,hi,ixs = bds.get(find(i),(int(-1e9),int(1e9),[]))
            ixs.append(i)
            bds[find(i)] = (max(a*q+b,lo),min(c*q+d,hi),ixs)
        bds = bds.values()
        for jxs in product(*[range(a,b+1) for a,b,ixs in bds]):
            kxs = [0 for i in range(len(abounds))]
            for j, ixs in zip(jxs,[ixs for _,_,ixs in bds]):
                for i in ixs:
                    kxs[i] = j
            kxs = tuple(kxs)
            T[kxs] = T.get(kxs,0) + val
    return T

# Performs the contraction operation on a tensor given sparesly as mappings from
# indices to values. For testing
def contract_explicit(Adict,i,j):
    out = {}
    i,j = sorted([i,j])
    for ixs,e in Adict.items():
        if ixs[i] == ixs[j]:
            jxs = ixs[:i]+ixs[i+1:j]+ixs[j+1:]
            out[jxs] = out.get(jxs,0) + e
    return out

# computes the multiplicity space maps for each representation of the koszul
# flattening map
# 
# representations are indicated by a pair of {0,1}
# with 0 representing the
# trivial representation V0 and 1 representing the standard representation V1
def square_matrices():
    q = QQ['q'].gen()

    cw = [ ]
    cw.append( (1,[(0,0,0,0), (0,1,1,0), (0,1,1,0)], [(1,2)] ) )
    cw.append( (1,[(0,1,1,0), (0,0,0,0), (0,1,1,0)], [(0,2)] ) )
    cw.append( (1,[(0,1,1,0), (0,1,1,0), (0,0,0,0)], [(0,1)] ) )

    projs = []
    projs.append( (1, [(0,0,0,0), (0,0,0,1), (0,0,0,1)], [ ]) )
    projs.append( (1, [(0,1,0,1), (0,0,1,0), (0,0,0,0)], [ ]) )
    projs.append( (1, [(0,1,0,1), (0,0,0,0), (0,1,1,0)], [ ]) )
    projs.append( constant_basic([2,0,3],1) )
    projs.append( constant_basic([2,2,2],1) )
    projs.append( constant_basic([2,3,0],1) )
    projs.append( constant_basic([2,3,1],1) )

    koszul_p1 = []
    koszul_p1.append(constant_basic([0,0,0],-1))
    koszul_p1.append(constant_basic([2,2,0],1))
    koszul_p1.append(constant_basic([0,1,1],1))
    koszul_p1.append(constant_basic([1,2,1],1))
    koszul_p1.append(constant_basic([1,0,2],-1))
    koszul_p1.append(constant_basic([2,1,2],-1))

    A0 = []
    A0.append( (1, [(0,0,0,3),(0,0,0,3)], [(0,1)]) )
    A0.append( (1, [ (0,4,0,4), (0,4,1,0) ], []) )

    A1 = []
    A1.append(constant_basic([0,4],-1))
    A1.append(constant_basic([0,5],1))

    B0 = []
    B0.append( (q-3, [(0,0,0,3), (0,0,0,3)], [(0,1)]) )
    B0.append( (1, [(0,4,1,0), (0,4,0,4)], []) )

    B1 = []
    B1.append(constant_basic([4,0],-(q-4)))
    B1.append( (1, [(0,5,1,0), (0,0,0,0)], []) )

    A = [A0,A1]
    B = [B0,B1]

    cw2 = transpose_box(tensor_product_box(cw,cw), [0,3,1,4,2,5])
    cw2Ap = contract_box(contract_box(tensor_product_box(projs, cw2),1,3),1,2)
    cw2Apwedge1 = transpose_box(contract_box( tensor_product_box(koszul_p1, cw2Ap),
        2,3), [0,2,3,1,4,5])

    fs = []
    for ixs in product(*[range(2) for i in range(2)]):
        trivmult = ixs.count(0)
        print ' '.join(['V%d' % i for i in ixs])
        AA = transpose_box(tensor_product_box(identity_tensor(3),
                reduce(tensor_product_box,[A[i] for i in ixs])),
                [0,2,4,1,3,5])
        BB = transpose_box(tensor_product_box(identity_tensor(3),
                reduce(tensor_product_box,[B[i] for i in ixs])),
                [0,2,4,1,3,5])
        AT = contract_box(contract_box(contract_box(tensor_product_box(
            AA,cw2Apwedge1), 3,6),3,5),3,4)
        ATB = contract_box(contract_box(contract_box(tensor_product_box(
            AT,BB),3,6),3,5),3,4)

        M = get_explicit(ATB,0,sub=False)
        mults = [5 if i==0 else 1 for i in ixs]
        Md = { (sum( prod(mults[i:])*ix for i,ix in enumerate(ixs[:3]) ),
                sum( prod(mults[i:])*ix for i,ix in enumerate(ixs[3:]) )) : e
                 for ixs,e in M.items()}

        Me = matrix(QQ['q'],3*5**trivmult,3*5**trivmult,Md)
        fs.append((ixs, (Me/(q-3)^2).change_ring(QQ['q']) ))

    return fs

def cube_matrices():
    q = QQ['q'].gen()

    cw = [ ]
    cw.append( (1,[(0,0,0,0), (0,1,1,0), (0,1,1,0)], [(1,2)] ) )
    cw.append( (1,[(0,1,1,0), (0,0,0,0), (0,1,1,0)], [(0,2)] ) )
    cw.append( (1,[(0,1,1,0), (0,1,1,0), (0,0,0,0)], [(0,1)] ) )

    projs = []
    projs.append(constant_basic([0,0,0,0],1))
    projs.append( (1,[(0,1,0,1), (0,1,1,0), (0,0,0,0), (0,0,0,0)], [] ) )
    projs.append( (1,[(0,1,0,1), (0,0,0,0), (0,1,1,0), (0,0,0,0)], [] ) )
    projs.append( (1,[(0,1,0,1), (0,0,0,0), (0,0,0,0), (0,1,1,0)], [] ) )
    projs.append(constant_basic([2,0,0,1], 1))
    projs.append(constant_basic([2,0,1,0], 1))
    projs.append(constant_basic([2,0,1,2], 1))
    projs.append(constant_basic([2,1,0,2], 1))
    projs.append(constant_basic([2,1,1,0], 1))
    projs.append(constant_basic([2,1,2,1], 1))
    projs.append(constant_basic([2,2,0,0], 1))
    projs.append(constant_basic([2,2,1,1], 1))
    projs.append(constant_basic([3,0,2,2], 1))
    projs.append(constant_basic([3,0,3,0], 1))
    projs.append(constant_basic([3,0,3,1], 1))
    projs.append(constant_basic([3,1,0,0], 1))
    projs.append(constant_basic([3,1,0,3], 1))
    projs.append(constant_basic([3,1,2,0], -1))
    projs.append(constant_basic([3,2,1,0], 1))
    projs.append(constant_basic([3,2,1,2], 1))
    projs.append(constant_basic([3,3,0,0], 1))
    projs.append(constant_basic([4,0,0,2], 1))
    projs.append(constant_basic([4,0,0,4], 1))
    projs.append(constant_basic([4,0,1,1], 1))
    projs.append(constant_basic([4,0,1,4], 1))
    projs.append(constant_basic([4,0,2,0], 1))
    projs.append(constant_basic([4,0,2,3], 1))
    projs.append(constant_basic([4,0,3,2], 1))
    projs.append(constant_basic([4,0,4,0], 1))
    projs.append(constant_basic([4,1,0,0], 1))
    projs.append(constant_basic([4,1,2,2], 1))
    projs.append(constant_basic([4,2,2,0], 1))
    projs.append(constant_basic([4,3,0,3], 1))

    koszul_p2 = []
    koszul_p2.append(constant_basic([0,2,0],1))
    koszul_p2.append(constant_basic([1,4,0],1))
    koszul_p2.append(constant_basic([2,5,0],1))
    koszul_p2.append(constant_basic([4,7,0],1))
    koszul_p2.append(constant_basic([5,8,0],1))
    koszul_p2.append(constant_basic([7,9,0],1))
    koszul_p2.append(constant_basic([0,1,1],-1))
    koszul_p2.append(constant_basic([1,3,1],-1))
    koszul_p2.append(constant_basic([3,5,1],1))
    koszul_p2.append(constant_basic([4,6,1],-1))
    koszul_p2.append(constant_basic([6,8,1],1))
    koszul_p2.append(constant_basic([8,9,1],1))
    koszul_p2.append(constant_basic([0,0,2],1))
    koszul_p2.append(constant_basic([2,3,2],-1))
    koszul_p2.append(constant_basic([3,4,2],-1))
    koszul_p2.append(constant_basic([5,6,2],-1))
    koszul_p2.append(constant_basic([6,7,2],-1))
    koszul_p2.append(constant_basic([9,9,2],1))
    koszul_p2.append(constant_basic([1,0,3],1))
    koszul_p2.append(constant_basic([2,1,3],1))
    koszul_p2.append(constant_basic([3,2,3],1))
    koszul_p2.append(constant_basic([7,6,3],-1))
    koszul_p2.append(constant_basic([8,7,3],-1))
    koszul_p2.append(constant_basic([9,8,3],-1))
    koszul_p2.append(constant_basic([4,0,4],1))
    koszul_p2.append(constant_basic([5,1,4],1))
    koszul_p2.append(constant_basic([6,2,4],1))
    koszul_p2.append(constant_basic([7,3,4],1))
    koszul_p2.append(constant_basic([8,4,4],1))
    koszul_p2.append(constant_basic([9,5,4],1))

    A0 = []
    A0.append( (1, [(0,0,0,4),(0,0,0,4)], [(0,1)]) )
    A0.append( (1, [ (0,5,0,5), (0,5,1,0) ], []) )

    A1 = []
    A1.append(constant_basic([0,5],-1))
    A1.append(constant_basic([0,6],1))

    B0 = []
    B0.append( (q-4, [(0,0,0,4), (0,0,0,4)], [(0,1)]) )
    B0.append( (1, [(0,5,1,0), (0,5,0,5)], []) )

    B1 = []
    B1.append(constant_basic([5,0],-(q-5)))
    B1.append( (1, [(0,6,1,0), (0,0,0,0)], []) )

    A = [A0,A1]
    B = [B0,B1]

    cw3 = transpose_box(tensor_product_box(cw,tensor_product_box(cw,cw)),
            [0,3,6,1,4,7,2,5,8])
    cw3Ap = contract_box(contract_box(contract_box(tensor_product_box(projs, 
        cw3),1,4),1,3),1,2)
    cw3Apwedge2 = transpose_box(contract_box( tensor_product_box(koszul_p2, cw3Ap),
        2,3), [0,2,3,4,1,5,6,7])

    fs = []
    for ixs in product(*[range(2) for i in range(3)]):
        trivmult = ixs.count(0)

        print ' '.join(['V%d' % i for i in ixs])
        AA = transpose_box(tensor_product_box(identity_tensor(10),
                reduce(tensor_product_box,[A[i] for i in ixs])),
                [0,2,4,6,1,3,5,7])
        BB = transpose_box(tensor_product_box(identity_tensor(10),
                reduce(tensor_product_box,[B[i] for i in ixs])),
                [0,2,4,6,1,3,5,7])
        AT = contract_box(contract_box(contract_box(contract_box(tensor_product_box(
            AA,cw3Apwedge2), 4,8),4,7),4,6),4,5) 
        ATB = contract_box(contract_box(contract_box(contract_box(tensor_product_box(
            AT,BB),4,8),4,7),4,6),4,5)

        M = get_explicit(ATB,0,sub=False)
        mults = [6 if i==0 else 1 for i in ixs]
        Md = { (sum( prod(mults[i:])*ix for i,ix in enumerate(ixs[:4]) ),
                sum( prod(mults[i:])*ix for i,ix in enumerate(ixs[4:]) )) : e
                 for ixs,e in M.items()}

        Me = matrix(QQ['q'],10*6**trivmult,10*6**trivmult,Md)
        fs.append((ixs, (Me/(q-4)^3).change_ring(QQ['q'])))

    return fs

#Ranks here
#The following two routines verify the ranks by providing, in each case, for the matrix m whose
#rank we want to compute, rectangular matrices, p,q, such that pmq is upper-triangular with +/- on diagonal

def check_square():
    print 'checking multiplicity map ranks for the proof of the square\n'
    print 'first calculating the multiplicity map in the box parameterized framework'
    print 'q lower bounds for which steps of the computation is valid will be printed\n'
    squares = dict(square_matrices())
    print '\nchecking now that the ranks are as required\n'
    factorizations = sage_eval(open('cw_square.txt').read(),locals={'q':QQ['q'].gen()})
    for rep, p,m,q in factorizations:
        p = matrix(FractionField(QQ['q']),*p)
        m = matrix(FractionField(QQ['q']),*m)
        q = matrix(FractionField(QQ['q']),*q)

        # m is the multiplicity map
        assert m == squares[rep]

        # p and q are matrices which upper triangularizes m and puts +-1 on the diagonal
        t = p*m*q
        assert all(i <= j for i,j in t.nonzero_positions())
        assert all(e in [1,-1] for e in t.diagonal())

        print 'multiplicity map rank for representation '+' '.join(['V%d' % i for i
            in rep]) + ' is >= %d' % t.dimensions()[0],

        # p has entries in the rational function field QQ(q). The least
        # common denominator of their denominators are the following (so that
        # the result is proved for any q not a root)
        print 'provided q is not a root of', p.denominator().factor()


def check_cube():
    print 'checking multiplicity map ranks for the proof of the cube\n'
    print 'first calculating the multiplicity map in the box parameterized framework'
    print 'q lower bounds for which steps of the computation is valid will be printed\n'
    cubes = dict(cube_matrices())
    print '\nchecking now that the ranks are as required\n'
    factorizations = sage_eval(open('cw_cube.txt').read(),locals={'q':QQ['q'].gen()})
    for rep, p,m,q in factorizations:
        p = matrix(FractionField(QQ['q']),*p)
        m = matrix(FractionField(QQ['q']),*m)
        q = matrix(FractionField(QQ['q']),*q)

        # m is the multiplicity map
        assert m == cubes[rep]

        # p and q are matrices which upper triangularizes m and puts +-1 on the diagonal
        t = p*m*q
        assert all(i <= j for i,j in t.nonzero_positions())
        assert all(e in [1,-1] for e in t.diagonal())

        print 'multiplicity map rank for representation '+' '.join(['V%d' % i for i
            in rep]) + ' is >= %d' % t.dimensions()[0],

        # p has entries in the rational function field QQ(q). The least
        # common denominator of their denominators are the following (so that
        # the result is proved for any q not a root)
        print 'provided q is not a root of', p.denominator().factor()

try:
    # little hack to make this block run only when executed from 
    #  the command line, and not when loaded in a sage interactive shell
    __file__
    check_square()
    print '\n\n'
    check_cube()
except NameError:
    pass

# vim: ft=python

